﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Dynamic;
using System.Linq.Expressions;
using System.IO;
using System.Net;
using System.Xml.Linq;
using System.Web;

namespace _6_DynamicObject
{
    /// <summary>
    /// Dynamic object for executing REST queries to arbitrary endpoints
    /// </summary>
    public class RestQuery : DynamicObject
    {
        private string segments;


        public string FileExtension { get; set; }
        public char IdentifierSeparator { get; set; }
        public string UserName { get; set; }
        public string Password { get; set; }
        public string BaseUrl { get; set; }

        /// <summary>
        /// Add a new segment to the url
        /// </summary>
        /// <param name="binder">Method and parameter names</param>
        /// <param name="args">Arguments passed</param>
        /// <param name="result">Return value from action</param>
        /// <returns>True if handled successfully</returns>
        public override bool TryGetMember(GetMemberBinder binder, out object result)
        {
            segments += binder.Name.ToLower() + "/";
            result = this;
            return true;
        }

        /// <summary>
        /// Execute REST query - caller can specify any method and parameters they need
        /// </summary>
        /// <param name="binder">Method and parameter names</param>
        /// <param name="args">Arguments passed</param>
        /// <param name="result">Return value from action</param>
        /// <returns>True if handled successfully</returns>
        public override bool TryInvokeMember(InvokeMemberBinder binder, object[] args, out object result)
        {
            Console.WriteLine("Method Name: " + binder.Name);
            Console.WriteLine("Argument Count: " + binder.CallInfo.ArgumentCount);
            Console.WriteLine("Argument Names:");

            foreach (var name in binder.CallInfo.ArgumentNames)
            {
                Console.WriteLine("  " + name);
            }

            Console.WriteLine("args:");
            foreach (var arg in args)
            {
                Console.WriteLine("Value: " + arg.ToString());
            }

            var queryParamsList = new List<string>();

            var argNames = binder.CallInfo.ArgumentNames;
            for (int i = 0; i < argNames.Count; i++)
            {
                queryParamsList.Add(argNames[i] + "=" + HttpUtility.UrlEncode(args[i] as string));
            }

            var casedID = binder.Name.Replace("Get", string.Empty);
            string delimitedID = TransformIdentifier(casedID);

            var queryParams =
                queryParamsList.Count > 0 ?
                    "?" + string.Join("&", queryParamsList.ToArray()) :
                    string.Empty;

            string url = 
                BaseUrl.TrimEnd('/') + "/" + 
                segments +
                delimitedID + 
                "." + FileExtension +
                queryParams;

            result = PerformQuery(url);

            // criteria for whether the call was correct
            return true;
        }

        /// <summary>
        /// Delimits a cased identifier by separating
        /// words with delimeter where capitalization occurs.
        /// i.e. DirectMessage becomes direct_message
        /// </summary>
        /// <param name="casedID"></param>
        /// <returns></returns>
        private string TransformIdentifier(string casedID)
        {
            var newString = new StringBuilder();

            for (int i = 0; i < casedID.Length; i++)
            {
                if (i != 0 && Char.IsUpper(casedID[i]))
                {
                    newString.Append(IdentifierSeparator.ToString() + Char.ToLower(casedID[i]));
                }
                else
                {
                    newString.Append(Char.ToLower(casedID[i]));
                }
            }

            return newString.ToString();
        }

        /// <summary>
        /// makes HTTP call to REST endpoint
        /// </summary>
        /// <param name="url">URL with all query info</param>
        /// <returns>List of objects to return</returns>
        public string PerformQuery(string url)
        {
            var req = HttpWebRequest.Create(url) as HttpWebRequest;

            req.Credentials = new NetworkCredential(UserName, Password);

            string responseXml = string.Empty;
 
            using (var resp = req.GetResponse())
            {
                StringReader txtRdr;

                using (var strm = resp.GetResponseStream())
                {
                    var strmRdr = new StreamReader(strm);
                    txtRdr = new StringReader(strmRdr.ReadToEnd());
                }

                responseXml = txtRdr.ReadToEnd();
            }

            return responseXml;
        }
    }
}
